/* $Id: fsend.cpp,v 1.5 2004/12/20 16:23:54 roca Exp $ */
/*
 *  Copyright (c) 2003-2004 INRIA - All rights reserved
 *  main authors: Christoph Neumann - christoph.neumann@inrialpes.fr
 *		  Vincent Roca - vincent.roca@inrialpes.fr
 *		  Julien Laboure - julien.laboure@inrialpes.fr
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */


/*
 * fsend.c
 *
 *	flute sender side functions
 */
#include "flute.h"


/*
 * Local variables
 */

/* available FEC codecs and associated information */
typedef struct fec_codec_info {
	int	available;		/* boolean */
	int	max_block_size;		/* raw value as defined by MCL */
	int	max_fragment_size;	/* big files are fragmented into */
					/* fragments of this size */
} fec_codec_info_t;

fec_codec_info_t	NULL_FEC_info;
fec_codec_info_t	RSE_FEC_info;
fec_codec_info_t	LDGM_STAIRCASE_FEC_info;
fec_codec_info_t	LDGM_TRIANGLE_FEC_info;



#if 0
/*
 * Local variables
 */
int fec_code = MCL_FEC_SCHEME_NULL; /* FEC code that will be used by the sender */
	/* XXX: The current version uses a single FEC code for the whole */
	/* session even if FEC can be specified on a per-object basis. */
int max_fragment_size;	/* Big files are fragmented into fragments of this
			   size */
#endif
unsigned int	toiindex = 0;
XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument*	fdtinstance;


void FluteSend (void)
{
	//int mcl_max_block_size; /* maximum size of a source block in bytes; */
				/* depends on the FEC codec used and its */
				/* limitations on the "k" parameter */
	int	Bytes_sent = 0;
	int	mcl_option;

	flute_lock(&flutemutex);
	fdtinstance = createNewFDTinstance();
	flute_unlock(&flutemutex);

	/* Determine available FEC codes and init. associated info struct */
	FindAvailableFEC();
	/* Determine the appropriate modes */
	/*
	 * NB: always use LCT1 now, in all cases...
	 */
	mcl_option = MCL_SCHED_LCT1;
	if (mcl_ctl(id, MCL_OPT_SCHED, (void*)&mcl_option, sizeof(mcl_option))) {
		EXIT(("Flute: mcl_ctl MCL_OPT_SCHED failed for LCT1\n"))
	}
#if 0
	/*
	 * Determine the maximum block size.
	 * This maximum is defined by the MCL library, depending on
	 * the FEC code in use, but it must not exceed the value
	 * specified in the * XXX_MAX_FRAGMENT_SIZE constant.
	 */
	if (mcl_ctl(id, MCL_OPT_GET_MAX_BLOCK_SIZE_FOR_CURRENT_FEC,
		    (void*)&mcl_max_block_size, sizeof(mcl_max_block_size))) {
		EXIT(("Flute: ERROR, mcl_ctl failed for MCL_OPT_GET_MAX_BLOCK_SIZE_FOR_CURRENT_FEC\n"))
	}
	ASSERT(mcl_max_block_size > 0);
	switch (fec_code) {
	case MCL_FEC_CODE_RSE:
		max_fragment_size = min(RSE_MAX_FRAGMENT_SIZE,
					mcl_max_block_size - MAX_TRAILER_SIZE);
		break;
#ifdef ALC
	case MCL_FEC_CODE_LDGM:
		max_fragment_size = min(LDPC_MAX_FRAGMENT_SIZE,
					mcl_max_block_size - MAX_TRAILER_SIZE);
		break;
#endif /* RM_PROTOCOL */
	default:
		/* no FEC encoding... */
		max_fragment_size = min(NO_FEC_MAX_FRAGMENT_SIZE,
					mcl_max_block_size - MAX_TRAILER_SIZE);
		break;
	}
#else
	/*
	 * FLUTE needs to send all files as a single object, no
	 * matter how large they are. Just define a max size since we have
	 * practical limitations on the buffer size we can allocate for a
	 * single large object.
	 */
	//max_fragment_size = FLUTE_MAX_FILE_SIZE;
#endif
	if (optimode == OPTIMIZE_SPACE) {
		mcl_option = MCL_SCHED_PARTIALLY_MIXED_ORDER;
		if (mcl_ctl(id, MCL_OPT_OBJ_SCHED, (void*)&mcl_option,
				sizeof(mcl_option)))
			EXIT(("Flute: mcl_ctl MCL_OPT_SCHED failed\n"))
	} else if (optimode == OPTIMIZE_SPEED) {
		mcl_option = MCL_SCHED_MIXED_ORDER;
		if (mcl_ctl(id, MCL_OPT_OBJ_SCHED, (void*)&mcl_option, sizeof(mcl_option)))
			EXIT(("Flute: mcl_ctl MCL_OPT_OBJ_SCHED failed\n"))
	} else if (optimode == OPTIMIZE_CPU) {
		mcl_option = MCL_SCHED_MIXED_ORDER;
		if (mcl_ctl(id, MCL_OPT_OBJ_SCHED, (void*)&mcl_option, sizeof(mcl_option)))
			EXIT(("Flute: mcl_ctl MCL_OPT_OBJ_SCHED failed\n"))
	} else {
		EXIT(("Flute: ERROR, invalid optimization mode!"))
	}
	/* NONEWADU is not FLUTE compatible, so don't use */
	if (mcl_ctl(id, MCL_OPT_NO_NONEWADU,NULL, 0)) {
		EXIT(("Flute: mcl_ctl MCL_OPT_NO_NONEWADU failed\n"))
	}

	/*
	mcl_option = 1;
	mcl_ctl(id, MCL_OPT_REUSE_APPLI_TX_BUFFER, (void*)&mcl_option,
		sizeof(mcl_option));
	*/
	/*
	 * Now send the file(s)...
	 */
	if(mcl_ctl(id, MCL_OPT_KEEP_DATA, NULL, 0))
		EXIT(("Flute: mcl_ctl KEEP_DATA failed\n"))
	if (recursive) {
		Bytes_sent = RecursiveSend(fileparam);
	} else {
		Bytes_sent = SendThisFile(fileparam);
	}
	/*
	 * ...and the associated FDT.
	 */
	char	*fdtbuffer;
	int	fdtsize = 0;
	int	mcl_option_toi = 0;	// set TOI to 0 for the FDT
	flute_lock(&flutemutex);
	
	/*fdtbuffer is allocated in getFinalFDTInstance*/
	fdtsize = getFinalFDTInstance(fdtinstance, &fdtbuffer);
	if (mcl_ctl(id, MCL_OPT_SET_NEXT_TOI, (void*)&mcl_option_toi,
		sizeof(mcl_option_toi))) {
		EXIT(("Flute: mcl_ctl MCL_OPT_SET_NEXT_TOI failed\n"))
	}
	if (mcl_ctl(id,  MCL_OPT_SET_NEXT_ADU_HIGH_IMPORTANCE, (void*)NULL, 0)) {
		EXIT(("Flute: mcl_ctl MCL_OPT_SET_NEXT_TOI failed\n"))
	}
	ChooseFEC(fdtsize);
	if (mcl_send(id, fdtbuffer , fdtsize) < 0)
		EXIT(("Flute: mcl_send failed\n"))
	
#ifdef WIN32
	/* display the FDT */
	displayFDT(0, NULL);
#endif
		
	flute_unlock(&flutemutex);
	if(mcl_ctl(id, MCL_OPT_PUSH_DATA, NULL, 0))
		EXIT(("Flute: mcl_ctl PUSH_DATA failed\n"))
	/*
	 * Finished, close the MCL session.
	 */
	mcl_close(id);
	PRINT(("\nFluteSend complete. %d bytes sent\n", Bytes_sent))
}


#ifdef WIN32

int RecursiveSend (char* Path)
{

	HANDLE			dirp = NULL;
	WIN32_FIND_DATA entry;

	unsigned int total_sent = 0;
	char FullName[MAX_PATH + MAX_FILENAME];
	char FindString[MAX_PATH];
	strcpy(FindString, Path);
	strcat(FindString, "\\*");
	
	dirp = FindFirstFile(FindString, &entry);

	if (dirp == INVALID_HANDLE_VALUE)
	{
		EXIT(("Flute: ERROR, in recursive mode, the given parameter MUST BE a valid directory name\nAborting...\n"))
	}

	if (!IsDirDots (entry.cFileName))
	{
		strcpy(FullName, Path);
		if(FullName[strlen(FullName)-1] == '\\')
			FullName[strlen(FullName)-1] = '/';

		if(FullName[strlen(FullName)-1] != '/')
			strcat(FullName,"/");

		strcat(FullName, entry.cFileName);

		if( entry.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
		{
			ASSERT((strcmp(Path,FullName)))
			PRINT(("Entering Directory %s\n", FullName))
			total_sent+= RecursiveSend (FullName);
		}
		else
		{
			PRINT(("\nSending File %s\n", FullName))
			total_sent += SendThisFile (FullName);
		}
	}

	while ( FindNextFile(dirp, &entry) )
	{
		if (IsDirDots (entry.cFileName))
			continue;

		strcpy(FullName, Path);
		if(FullName[strlen(FullName)-1] == '\\')
			FullName[strlen(FullName)-1] = '/';

		if(FullName[strlen(FullName)-1] != '/')
			strcat(FullName,"/");

		strcat(FullName, entry.cFileName);

		if( entry.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
		{
			ASSERT((strcmp(Path,FullName)))
			PRINT(("Entering Directory %s\n", FullName))
			total_sent+= RecursiveSend (FullName);
		}
		else
		{
			PRINT(("\nSending File %s\n", FullName))
			total_sent += SendThisFile (FullName);
		}
	}
	FindClose(dirp);
	return total_sent;
}


#else

int RecursiveSend (char* Path)
{
	unsigned int total_sent = 0;
	struct dirent *entry;
	struct stat stats;
	char FullName[MAX_PATH + MAX_FILENAME];
	DIR *dirp;

	if ((dirp = opendir (Path)) == NULL)
	{
		perror(Path);
		EXIT(("RecursiveSend: ERROR, opendir failed\n"))
	}

	/*
	For SOLARIS users:
	Why doesn't readdir work? It chops the first two characters of
	all filenames.

	You're probably linking with libucb and didn't read question
	6.18. (Readdir in libucb.so wants you to include sys/dir.h,
	but many SunOS 4.1.x programs included <dirent.h>, consequently,
	you're mixing native <dirent.h> struct dirent with libucb
	readdir(). The symptom of this mixup is that the first two
	characters of each filename are missing. Make sure you use a
	native compiler (default /opt/SUNWspro/bin/cc, which may not be in
	your PATH), and not /usr/ucb/cc. 
	*/

	while ((entry = readdir(dirp)) != NULL) {
		if (IsDirDots (entry->d_name))
			continue;

		strcpy(FullName, Path);
		if (FullName[strlen(FullName)-1] != '/')
			strcat(FullName, "/");

		strcat(FullName, entry->d_name);

		if (stat(FullName, &stats) == -1) {
			perror("RecursiveSend: ERROR, stat failed");
			PRINT(("RecursiveSend: FullName=%s\nentry->d_name=%s\nPath=%s\n", FullName, entry->d_name, Path))
			EXIT(("RecursiveSend: ERROR, stat() failed\n"))
		}

		if (S_ISDIR(stats.st_mode)) {
			ASSERT((strcmp(Path,FullName)))
			PRINT(("Entering Directory %s\n", FullName))
			total_sent+= RecursiveSend (FullName);
		} else {
			PRINT(("Sending File %s\n", FullName))
			total_sent += SendThisFile (FullName);
		}
	}
	closedir (dirp);
	return total_sent;
}

#endif



int SendThisFile (char *file_path) // int fragment_len)
{
	fec_codec_info_t *codec_info;
	struct		stat file_stats;
	int		file_to_send 	= 0;
	char		*buf_file	= NULL;
	int		sent		= 0;
	int		ObjectLength 	= 0;
#define TEMP_STRING_LEN	128
	char 		*temp_string;
	int		len;
#ifndef WIN32
	unsigned char	*md5sum;
#endif
	XERCES_CPP_NAMESPACE_QUALIFIER DOMElement	*el;

	if ((temp_string = (char*)malloc(TEMP_STRING_LEN)) == NULL) {
		EXIT(("Flute: SendThisFile: Error, malloc failed\n"))
	}
	/*
	 * Set the TOI for this file
	 */
	toiindex++;
	if (mcl_ctl(id, MCL_OPT_SET_NEXT_TOI, (void*)&toiindex,
			sizeof(toiindex)))
		EXIT(("Flute: mcl_ctl MCL_OPT_SET_NEXT_TOI failed\n"))

	if (FileExist(file_path)) {
		if (stat(file_path, &file_stats) == -1) {
			EXIT(("Flute: SendThisFile: Error: stat()\n"))
			//PRINT(("Flute: SendThisFile: Error: stat()\n"))
			//goto end;
		}
#ifdef WIN32
		file_to_send = open(file_path, O_RDONLY | O_BINARY);
#else
		file_to_send = open(file_path, O_RDONLY);
#endif
		if (file_to_send < 0) {
			EXIT(("Flute: open failed for file %s\n", file_path))
		}
	} else {
		PRINT(("Flute: SendThisFile: Error: %s, no such file!\n", file_path))
		goto end;
	}
	if (!(file_stats.st_mode & S_IFREG))
		EXIT(("Flute: SendThisFile: Error: %s is not a regular file\n", file_path))		

	codec_info = ChooseFEC((int)(file_stats.st_size));
	if (!(buf_file = (char*)malloc(file_stats.st_size))) {
		EXIT(("Flute: SendThisFile: Error: out of memory (%d bytes asked)!\n", file_stats.st_size))
	}

	if ((ObjectLength = read(file_to_send, buf_file, file_stats.st_size)) < 0)
		EXIT(("Flute: SendThisFile: Error FluteSend: fread failed, returned %d\n", ObjectLength))

	if (mcl_send(id, buf_file, ObjectLength) < 0)
		EXIT(("Flute: SendThisFile: mcl_send failed\n"))
	sent+= ObjectLength;
	free(buf_file);

	/*
	 * Update the FDT now...
	 */
	flute_lock(&flutemutex);

	el=createNewFile();

	/*TOI*/
#ifdef WIN32
	len = sprintf(temp_string, "%u", toiindex);
#else
	len = snprintf(temp_string, TEMP_STRING_LEN, "%u", toiindex);
#endif
	if (len < 0 || len >= TEMP_STRING_LEN) {
		EXIT(("Flute: SendThisFile: snprintf failed, returned %d)\n", len))
	}
	el=setFileAttribute(el, "TOI", temp_string);

	/*Content-Location*/
	el=setFileAttribute(el, "Content-Location", file_path);

	/*Content-Length*/
#ifdef WIN32
	len = sprintf(temp_string, "%lu", file_stats.st_size);
#else
	len = snprintf(temp_string, TEMP_STRING_LEN, "%lu", file_stats.st_size);
#endif
	if (len < 0 || len >= TEMP_STRING_LEN) {
		EXIT(("Flute: SendThisFile: snprintf failed, returned %d)\n", len))
	}
	el=setFileAttribute(el, "Content-Length", temp_string);

#ifndef WIN32
	/*Content-MD5*/
	if ((md5sum = (unsigned char*)calloc(1, MD5BASE64_LENGTH)) == NULL) {
		EXIT(("Flute: SendThisFile: Error, calloc failed"))
	}
	md5sum_calculate(file_to_send, &md5sum);	
	el=setFileAttribute(el, "Content-MD5", (char *)md5sum);
	free(md5sum);
#endif

	appendFile(el);

	/* Adding to fdt-instance*/
	AddFileToFDTinstance(fdtinstance, toiindex);

	flute_unlock(&flutemutex);

	close(file_to_send);

end:
	free(temp_string);
	return sent;
}


/*
 * Updates the FEC available global variables depending on the available
 * FEC codecs.
 * Sets the FEC ratio for each possible FEC codec.
 * Retrieves the maximum block size for each possible FEC codec.
 */
void
FindAvailableFEC (void)
{
	int	codec;	/* temp variable */
	int	max_sz;	/* temp variable */

	ASSERT(fec_ratio >= 1.0);
	/*
	 * NULL_FEC codec first (should always be supported)
	 */
	codec = MCL_FEC_SCHEME_NULL;
	if (mcl_ctl(id, MCL_OPT_SET_FEC_CODE, (void*)&codec, sizeof(codec))) {
		EXIT(("Flute: ERROR, FEC codec NULL not available\n"))
	} else {
		NULL_FEC_info.available = 1;
		/*
		 * determine the max block size, as defined by MCL,
		 * and the associated max file fragment size, used
		 * by FCAST.
		 */
		if (mcl_ctl(id,
			    MCL_OPT_GET_MAX_BLOCK_SIZE_FOR_CURRENT_FEC,
			    (void*)&max_sz, sizeof(max_sz))) {
			EXIT(("Flute: ERROR, mcl_ctl failed for MCL_OPT_GET_MAX_BLOCK_SIZE_FOR_CURRENT_FEC for LDGM codec\n"))
		}
		NULL_FEC_info.max_block_size = max_sz;
		NULL_FEC_info.max_fragment_size =
						min(NULL_FEC_MAX_FRAGMENT_SIZE,
						    max_sz);
	}
	/*
	 * now let's see the other FEC codecs
	 */
	if (fec_ratio == 1.0) {
		/* user wants no FEC */
		RSE_FEC_info.available = 0;
		LDGM_STAIRCASE_FEC_info.available = 0;
		LDGM_TRIANGLE_FEC_info.available = 0;
	} else {
		/* user wants some FEC */
		/*
		 * RSE
		 */
		codec = MCL_FEC_SCHEME_RSE_129_0;
		if (mcl_ctl(id, MCL_OPT_SET_FEC_CODE, (void*)&codec,
				sizeof(codec))) {
			RSE_FEC_info.available = 0;
		} else {
			RSE_FEC_info.available = 1;
			if (mcl_ctl(id, MCL_OPT_FEC_RATIO, (void*)&fec_ratio,
					sizeof(fec_ratio))) {
				EXIT(("Flute: ERROR, mcl_ctl failed for FEC_RATIO %d and RSE\n", fec_ratio))
			}
			/*
			 * determine the max block size, as defined by MCL,
			 * and the associated max file fragment size, used
			 * by FCAST.
			 */
			if (mcl_ctl(id,
				    MCL_OPT_GET_MAX_BLOCK_SIZE_FOR_CURRENT_FEC,
				    (void*)&max_sz, sizeof(max_sz))) {
				EXIT(("Flute: ERROR, mcl_ctl failed for MCL_OPT_GET_MAX_BLOCK_SIZE_FOR_CURRENT_FEC for RSE codec\n"))
			}
			RSE_FEC_info.max_block_size = max_sz;
			RSE_FEC_info.max_fragment_size =
						min(RSE_MAX_FRAGMENT_SIZE,
						    max_sz);
			/* tests MCL blocking algo: define a large frag size */
			//RSE_FEC_info.max_fragment_size = 512*1024;
		}
#if defined(ALC)
		/*
		 * LDGM STAIRCASE
		 */
		codec = MCL_FEC_SCHEME_LDGM_STAIRCASE_132_0;
		if (mcl_ctl(id, MCL_OPT_SET_FEC_CODE, (void*)&codec,
				sizeof(codec))) {
			LDGM_STAIRCASE_FEC_info.available = 0;
		}else {
			LDGM_STAIRCASE_FEC_info.available = 1;
			if (mcl_ctl(id, MCL_OPT_FEC_RATIO, (void*)&fec_ratio,
					sizeof(fec_ratio))) {
				EXIT(("Flute: ERROR, mcl_ctl failed for FEC_RATIO %f for LDGM STAIRCASE codec\n", fec_ratio))
			}
			/*
			 * determine the max block size, as defined by MCL,
			 * and the associated max file fragment size, used
			 * by FCAST.
			 */
			if (mcl_ctl(id,
				    MCL_OPT_GET_MAX_BLOCK_SIZE_FOR_CURRENT_FEC,
				    (void*)&max_sz, sizeof(max_sz))) {
				EXIT(("Flute: ERROR, mcl_ctl failed for MCL_OPT_GET_MAX_BLOCK_SIZE_FOR_CURRENT_FEC for LDGM STAIRCASE codec\n"))
			}
			LDGM_STAIRCASE_FEC_info.max_block_size = max_sz;
			LDGM_STAIRCASE_FEC_info.max_fragment_size =
						min(LDGM_MAX_FRAGMENT_SIZE,
						    max_sz);
		}
		/*
		 * LDGM TRIANGLE
		 */
		codec = MCL_FEC_SCHEME_LDGM_TRIANGLE_132_1;
		if (mcl_ctl(id, MCL_OPT_SET_FEC_CODE, (void*)&codec,
				sizeof(codec))) {
			LDGM_TRIANGLE_FEC_info.available = 0;
		}else {
			LDGM_TRIANGLE_FEC_info.available = 1;
			if (mcl_ctl(id, MCL_OPT_FEC_RATIO, (void*)&fec_ratio,
					sizeof(fec_ratio))) {
				EXIT(("Flute: ERROR, mcl_ctl failed for FEC_RATIO %f for LDGM TRIANGLE codec\n", fec_ratio))
			}
			/*
			 * determine the max block size, as defined by MCL,
			 * and the associated max file fragment size, used
			 * by FCAST.
			 */
			if (mcl_ctl(id,
				    MCL_OPT_GET_MAX_BLOCK_SIZE_FOR_CURRENT_FEC,
				    (void*)&max_sz, sizeof(max_sz))) {
				EXIT(("Flute: ERROR, mcl_ctl failed for MCL_OPT_GET_MAX_BLOCK_SIZE_FOR_CURRENT_FEC for LDGM TRIANGLE codec\n"))
			}
			LDGM_TRIANGLE_FEC_info.max_block_size = max_sz;
			LDGM_TRIANGLE_FEC_info.max_fragment_size =
						min(LDGM_MAX_FRAGMENT_SIZE,
						    max_sz);
		}
#elif defined(NORM)
		LDGM_STAIRCASE_FEC_info.available = 0;
		LDGM_TRIANGLE_FEC_info.available = 0;
#endif /* RM_PROTOCOL */
		// Remove following two line comment to test with RSE
		//LDGM_STAIRCASE_FEC_info.available = 0;
		//LDGM_TRIANGLE_FEC_info.available = 0;
		if (!RSE_FEC_info.available && !LDGM_STAIRCASE_FEC_info.available && !LDGM_TRIANGLE_FEC_info.available) {
			EXIT(("Flute: ERROR, at least one of RSE or LDGM codecs must be available\n"))
		}
	}
}


/*
 * Choose the most appropriate FEC codec, depending on the file size
 * and the available FEC codecs.
 * @param file_size	file size
 * @return		FEC codec chosen
 */
fec_codec_info_t *
ChooseFEC (int	file_size)
{
	int		codec;
	fec_codec_info_t *codec_info;

	if (fec_ratio == 1.0) {
		ASSERT(NULL_FEC_info.available);
		codec = MCL_FEC_SCHEME_NULL;
		codec_info = &NULL_FEC_info;
		goto codec_found;
	}
	/*
	 * With ALC, use LDGM first for big files and RSE for
	 * small files, and if the default choice is not available,
	 * use the other one.
	 */
	if (file_size >= 10 * RSE_LDGM_FILE_SIZE_THRESHOLD) {
		if (fec_ratio >= 2.5 && LDGM_STAIRCASE_FEC_info.available) {
			/*
			 * Use LDGM Staircase with large FEC expansion ratios.
			 */
			codec = MCL_FEC_SCHEME_LDGM_STAIRCASE_132_0;
			codec_info = &LDGM_STAIRCASE_FEC_info;
		} else if (LDGM_TRIANGLE_FEC_info.available) {
			/*
			 * Use LDGM Triangle with small FEC expansion ratios.
			 */
			codec = MCL_FEC_SCHEME_LDGM_TRIANGLE_132_1;
			codec_info = &LDGM_TRIANGLE_FEC_info;
		} else {
			ASSERT(RSE_FEC_info.available)
			codec = MCL_FEC_SCHEME_RSE_129_0;
			codec_info = &RSE_FEC_info;
		}
	} else {
		if (RSE_FEC_info.available) {
			codec = MCL_FEC_SCHEME_RSE_129_0;
			codec_info = &RSE_FEC_info;
		} else {
			ASSERT(LDGM_STAIRCASE_FEC_info.available)
			codec = MCL_FEC_SCHEME_LDGM_STAIRCASE_132_0;
			codec_info = &LDGM_STAIRCASE_FEC_info;
		}
	}

codec_found:
	/* set the codec now */
	if (mcl_ctl(id, MCL_OPT_SET_FEC_CODE, (void*)&codec, sizeof(codec))) {
		EXIT(("Flute: ERROR, ctl for MCL_OPT_SET_FEC_CODE (%d) failed\n", codec))
	}
	return codec_info;
}


#if 0
/*
 * Determines the optimum file segmentation, and the appropriate FEC codec
 * to use.
 * @param
 * @return
 */
int
FileSegmentation
{
}
#endif
